import argparse
import os
import sys
from pathlib import Path as P

import numpy as np
import pandas as pd
from netCDF4 import Dataset as netcdf_dataset
from netCDF4 import date2num as netcdf_date2num
from renku.core.commands import dataset
from scipy.interpolate import RegularGridInterpolator

import pyantarctica.dataset as adata

# print(os.getcwd())
# Import local renku client anyways? -- MV20200427: does not seem necessary

def interpolate_track(input, output_file, ecmwf_raw, VERBOSE=False):

    if "fiveminute_average_crui_10" not in os.listdir("data/"):
        dataset.import_dataset(uri=input, commit_message="imported from Zenodo")

    track_file = P([pp.path for pp in dataset.list_files() if ".csv" in pp.path][0])

    # if output_file not in os.listdir("data/"): # Move at the final export
    #     dataset.create_dataset(output_file, commit_message='Created empty dataset')

    # specifiy here variables with 0.5° x 0.5° degrees grid size. Not included variables are on a 0.25° x 0.25°
    grid05 = ["p140209", "mdts", "mdww", "msqs", "pp1d", "tmax", "shts", "shww"]

    era_file_list = sorted(os.listdir(ecmwf_raw))  # list of all nc files

    f_r = era_file_list[0]  # look at the first file to collect necessary info

    era5 = netcdf_dataset(ecmwf_raw + f_r, "r")
    era5_time_units = era5.variables["time"].units
    era5_time_calendar = era5.variables["time"].calendar

    var_list = list(era5.variables.keys())[
        3:
    ]  # first three entries are longitude, latitude, time
    var_shape = np.shape(
        era5.variables[var_list[0]]
    )  # get the shape of the variables (24, 721, 1440)

    era5.close()

    if VERBOSE:
        print(era5_time_units)
        print(era5_time_calendar)

        print("Variables selected for interpolation")
        print(var_list)

    # read the 5min ship track on which the data will be interpolated
    # ship_track = adata.read_standard_dataframe(track_file, crop_legs=False)
    ship_track = pd.read_csv(track_file)
    ship_track = ship_track.set_index(
        pd.to_datetime(ship_track.date_time, format="%Y-%m-%dT%H:%M:%S")
    )
    ship_track.drop(columns=["date_time"], inplace=True)

    # ensure tz-naive
    if str(ship_track.index.tzinfo) == "None":
        print(
            "The time stamp is provided is tz naive. It is assumed that the times provided are in UTC"
        )
    else:
        print(
            "The provided time zone of the track is "
            + str(ship_track.index.tzinfo)
            + ". This is now converted to a tz naive time stampe in UTC"
        )
        ship_track.index = ship_track.index.tz_convert(None)

    # converte the pandas datetime index to era5 compatible hours since 1900-01-01 00:00:00.0
    hours = np.squeeze(
        netcdf_date2num(
            ship_track.index.to_pydatetime(), era5_time_units, era5_time_calendar
        )
    )
    longitudes = np.squeeze(ship_track["longitude"].values)
    longitudes = longitudes % 360  # ERA latitudes are mapped on 0-360 degree
    latitudes = np.squeeze(ship_track["latitude"].values)

    ### MAIN LOOP OVER NETCDF FILES AND INTERPOLATIONS
    # create two empty data frame to be filled with the interpolations
    # one for nearest, and one for linear interpolation
    era5_track_nearest = pd.DataFrame(index=ship_track.index, columns=var_list)
    era5_track_linear = pd.DataFrame(index=ship_track.index, columns=var_list)

    # loop through the era files 'f_date_*.nc'
    for j in np.arange(0, len(era_file_list)):
        f_r = era_file_list[j]
        #     clear_output()
        print(f"file {f_r}")

        era5 = netcdf_dataset(ecmwf_raw + f_r, "r")

        # build the grids [time, latitude, longitude]
        # -lat ! to confrome with RegularGridInterpolator, ascending rule
        # one grid for the full 0.25 gridded variables
        # one grid for the reduced 0.50 gridded variables

        if j > 0:
            # if we are passed the first file, reload it to concat with current, this way we avoid gaps at the file edges
            era5_previous = netcdf_dataset(ecmwf_raw + era_file_list[j - 1], "r")
            era5grid = [
                np.concatenate(
                    [era5_previous.variables["time"][:], era5.variables["time"][:]]
                ),
                -era5.variables["latitude"][:],
                era5.variables["longitude"][:],
            ]

            era5grid05deg = [
                np.concatenate(
                    [era5_previous.variables["time"][:], era5.variables["time"][:]]
                ),
                -era5.variables["latitude"][::2],
                era5.variables["longitude"][::2],
            ]

        else:

            era5grid = [
                era5.variables["time"][:],
                -era5.variables["latitude"][:],
                era5.variables["longitude"][:],
            ]

            era5grid05deg = [
                era5.variables["time"][:],
                -era5.variables["latitude"][::2],
                era5.variables["longitude"][::2],
            ]

        for var_str in var_list:
            # check if the variable has a propar mask array with it or if the mask is just zero dim
            var_has_mask = np.shape(era5.variables[var_str][:, :, :].mask) == np.shape(
                era5.variables[var_str][:, :, :].data
            )
            # interpolate variable to track
            for itp_method in ["nearest", "linear"]:  #'nearest'
                # seperate treatment for 0.25 and 0.50 degree grid
                if var_str in grid05:

                    if j > 0:
                        var05deg = np.concatenate(
                            [
                                era5_previous.variables[var_str][
                                    :,
                                    : 1 + int(np.floor(var_shape[1] / 2)),
                                    : int(np.floor(var_shape[2] / 2)),
                                ].data,
                                era5.variables[var_str][
                                    :,
                                    : 1 + int(np.floor(var_shape[1] / 2)),
                                    : int(np.floor(var_shape[2] / 2)),
                                ].data,
                            ]
                        )
                        if var_has_mask:
                            var05deg_mask = np.concatenate(
                                [
                                    era5_previous.variables[var_str][
                                        :,
                                        : 1 + int(np.floor(var_shape[1] / 2)),
                                        : int(np.floor(var_shape[2] / 2)),
                                    ].mask,
                                    era5.variables[var_str][
                                        :,
                                        : 1 + int(np.floor(var_shape[1] / 2)),
                                        : int(np.floor(var_shape[2] / 2)),
                                    ].mask,
                                ]
                            )
                            var05deg[
                                var05deg_mask == True
                            ] = np.NaN  # set masked data to NaN

                    else:
                        var05deg = era5.variables[var_str][
                            :,
                            : 1 + int(np.floor(var_shape[1] / 2)),
                            : int(np.floor(var_shape[2] / 2)),
                        ].data
                        if var_has_mask:
                            var05deg_mask = era5.variables[var_str][
                                :,
                                : 1 + int(np.floor(var_shape[1] / 2)),
                                : int(np.floor(var_shape[2] / 2)),
                            ].mask
                            var05deg[
                                var05deg_mask == True
                            ] = np.NaN  # set masked data to NaN

                    itp = RegularGridInterpolator(
                        era5grid05deg,
                        var05deg,
                        method=itp_method,
                        bounds_error=False,
                        fill_value=np.NaN,
                    )

                else:

                    if j > 0:
                        var = np.concatenate(
                            [
                                era5_previous.variables[var_str][:, :, :].data,
                                era5.variables[var_str][:, :, :].data,
                            ]
                        )
                        if var_has_mask:
                            var_mask = np.concatenate(
                                [
                                    era5_previous.variables[var_str][:, :, :].mask,
                                    era5.variables[var_str][:, :, :].mask,
                                ]
                            )
                            var[var_mask == True] = np.NaN  # set masked data to NaN
                    else:
                        var = era5.variables[var_str][:, :, :].data
                        if var_has_mask:
                            var_mask = era5.variables[var_str][:, :, :].mask
                            var[var_mask == True] = np.NaN  # set masked data to NaN

                    itp = RegularGridInterpolator(
                        era5grid,
                        var,
                        method=itp_method,
                        bounds_error=False,
                        fill_value=np.NaN,
                    )
                # interpolate the data onto the ship track
                # with var beeing a simple numpy arry in all cases we can now use itp the same way for both linear and nearest interpolation
                var_track = itp((hours, -latitudes, longitudes))
                # note the interpolated data from this nc file in the DataFrames
                if itp_method == "nearest":
                    era5_track_nearest.loc[~(np.isnan(var_track)), var_str] = var_track[
                        ~(np.isnan(var_track))
                    ]
                elif itp_method == "linear":
                    era5_track_linear.loc[~(np.isnan(var_track)), var_str] = var_track[
                        ~(np.isnan(var_track))
                    ]
        era5.close()
        if j > 0:
            era5_previous.close()

    # Rename with less cryptic column names
    rename = {
        "u10n": "10m_u_component_of_neutral_wind",
        "v10n": "10m_v_component_of_neutral_wind",
        "u10": "10m_u_component_of_wind",
        "v10": "10m_v_component_of_wind",
        "p140209": "air_density_over_the_oceans",
        "d2m": "2m_dewpoint_temperature",
        "t2m": "2m_temperature",
        "blh": "boundary_layer_height",
        "cbh": "cloud_base_height",
        "cp": "convective_precipitation",
        "csf": "convective_snowfall",
        "zust": "friction_velocity",
        "hcc": "high_cloud_cover",
        "lsm": "land_sea_mask",
        "lsp": "large_scale_precipitation",
        "lsf": "large_scale_snowfall",
        "lcc": "low_cloud_cover",
        "mdts": "mean_direction_of_total_swell",
        "mdww": "mean_direction_of_wind_waves",
        "msqs": "mean_square_slope_of_waves",
        "mslhf": "mean_surface_latent_heat_flux",
        "msshf": "mean_surface_sensible_heat_flux",
        "mcc": "medium_cloud_cover",
        "pp1d": "peak_wave_period",
        "tmax": "period_corresponding_to_maximum_individual_wave_height",
        "siconc": "sea_ice_cover",
        "sst": "sea_surface_temperature",
        "shts": "significant_height_of_total_swell",
        "shww": "significant_height_of_wind_waves",
        "skt": "skin_temperature",
        "tcc": "total_cloud_cover",
        "tciw": "total_column_cloud_ice_water",
        "tclw": "total_column_cloud_liquid_water",
    }
    # add latitude / longitude
    era5_track_linear["latitude"] = ship_track["latitude"]
    era5_track_linear["longitude"] = ship_track["longitude"]
    era5_track_nearest["latitude"] = ship_track["latitude"]
    era5_track_nearest["longitude"] = ship_track["longitude"]

    # rename the output columns
    era5_track_linear.rename(columns=rename, inplace=True)
    era5_track_nearest.rename(columns=rename, inplace=True)

    # Save tracks
    era5_track_linear.to_csv(
        "data/" + output_file + "/era5-on-cruise-track-5min-legs0-4-linear.csv",
        date_format="%Y-%m-%dT%H:%M:%S+00:00",
        na_rep="NaN",
    )  # //#

    era5_track_nearest.to_csv(
        "data/" + output_file + "/era5-on-cruise-track-5min-legs0-4-nearest.csv",
        date_format="%Y-%m-%dT%H:%M:%S+00:00",
        na_rep="NaN",
    )  # //#\

    # Create renku dataset object
    dataset.add_file(
        urls=["data/" + output_file + "/era5-on-cruise-track-5min-legs0-4-linear.csv"],
        short_name=output_file,
        create=True,
        sources=None,
        commit_message="added linear interpolation file",
    )

    dataset.add_file(
        urls=["data/" + output_file + "/era5-on-cruise-track-5min-legs0-4-nearest.csv"],
        short_name=output_file,
        create=False,
        sources=None,
        commit_message="added nearest neighbour interpolation file",
    )

    # Edit metadata
    title, description, creators = parse_readme(output_file)
    if title:
        dataset.edit_dataset(
            short_name=output_file,
            title=title,
            description=description,
            creators=creators,
            commit_message="Edited metadata",
        )


def parse_readme(output_file):
    fullpath = "data/" + output_file + "/" + "README.txt"

    if os.path.exists(fullpath):
        with open(fullpath, "r") as f:

            description = []
            append_desc = False

            for iline, line in enumerate(f.readlines()):
                if iline == 0:
                    title = line

                if line == "***** Dataset abstract *****\n":
                    append_desc = True

                if line == "***** Original data collection *****\n":
                    append_desc = False

                if append_desc:
                    description.append(line)

        description = description[2:]
        description = "".join(description)
        creators = "M. Volpi, S. Landwehr, J. Thomas, J. Schmale."
        creators = creators.split(", ")

        return title, description, creators
    else:
        print(f"{fullpath} missing.")
        return None, None, None


if __name__ == "__main__":

    apa = argparse.ArgumentParser()
    apa.format_help()
    apa.add_argument(
        "-i",
        "--input",
        required=False,
        help="Zenodo DOI of the input data",
        default="10.5281/zenodo.3772377", # 10.5281/zenodo.3772377 is v1.1


    )
    apa.add_argument(
        "-o",
        "--output",
        required=False,
        help="Output file name",
        default="ecmwf-era5-interpolated-to-cruise-track",
    )
    apa.add_argument(
        "-e",
        "--ecmwf-files",
        required=True,
        help="Path of the raw ECMWF ERA5 netCDF archives",
        default="data/ecmwf_raw/",
    )
    args = vars(apa.parse_args())

    interpolate_track(
        input=args["input"],
        output_file=args["output"],
        ecmwf_raw=args["ecmwf_files"],
    )